'use strict';
var Site = require('dw/system/Site');
var currentSite = Site.getCurrent();
var Logger = require('dw/system/Logger').getLogger('scalapay');
var Transaction = require('dw/system/Transaction');

function removeExistingPaymentInstruments(basket, method) {
    var iter = basket.getPaymentInstruments(method).iterator();

    // Remove payment instruments.
    while (iter.hasNext()) {
        basket.removePaymentInstrument(iter.next());
    }
}

// Calculate ScalaPay payment authorize service. If "scalapayDelayCaptureEnabled" set to true - payment capturing will be delayed
function getScalaPayPaymentAuthorizeService(paymentToken){
    var scalapayServices = require('*/cartridge/scripts/service/scalapayment');
    var isDelayCaptureEnabled = currentSite.getCustomPreferenceValue('scalapayDelayCaptureEnabled');
    var paymentAuthorizeService;

    if(isDelayCaptureEnabled){
        paymentAuthorizeService = scalapayServices.ScalaPayAuthorizeDelayService;
        var initialServiceURL = paymentAuthorizeService.getURL();
        var personalizedServiceURL = initialServiceURL.replace('{token}', paymentToken);
        paymentAuthorizeService.setURL(personalizedServiceURL);
    }else{
        paymentAuthorizeService = scalapayServices.ScalaPayAuthorizeService;
    }
    return paymentAuthorizeService;
}

// Calculate payment authorization request data based on "scalapayDelayCaptureEnabled" preference.
function getAuthorizationReqData(orderNumber, paymentToken){
    var isDelayCaptureEnabled = currentSite.getCustomPreferenceValue('scalapayDelayCaptureEnabled');
    var reqData = {
        merchantReference: orderNumber
    };

    if(!isDelayCaptureEnabled){
        reqData.token = paymentToken;
    }

    return reqData;
}

function handle(handlePaymentObj) {
    var basket = handlePaymentObj.Basket;
    var scalapayHelper = require('*/cartridge/scripts/helpers/scalapayHelper');
    var URLUtils = require('dw/web/URLUtils');
    var isScalapayApplicable = scalapayHelper.isPaymentApplicable(basket);
    var scalapayOrderToken = scalapayHelper.getPaymentToken(basket);
    var handlePaymentStatus = {
        error: true
    };

    // no actions if payment authorized
    if (isScalapayApplicable && session.privacy.scalapayAuthURL && scalapayOrderToken && scalapayHelper.isPaymentDataSync(basket)) {
        handlePaymentStatus.error = false;
        handlePaymentStatus.redirectURL = URLUtils.https('SCALAPAY-HandlePayment', 'status', 'success', 'orderToken', scalapayOrderToken).toString();
        return handlePaymentStatus;
    }

    if (isScalapayApplicable) {
        var handlePaymentData = scalapayHelper.getPaymentReqData(basket),
            scalapayService = require('*/cartridge/scripts/service/scalapayment'),
            scalapayUtils = require('*/cartridge/scripts/helpers/scalapayUtils'),
            authToken = currentSite.getCustomPreferenceValue('scalaPayPaymentToken'),
            authAmountMoney = scalapayUtils.getNonGiftCertificateAmount(basket);

        Transaction.wrap(function () {
            removeExistingPaymentInstruments(basket, 'CREDIT_CARD');
            removeExistingPaymentInstruments(basket, 'ScalaPay');
            basket.createPaymentInstrument('ScalaPay', authAmountMoney);
        });

        if (authToken) {
            session.privacy.scalapayAuthURL = null;
            var handlePaymentResult = scalapayService.ScalaPayHandleService.call(handlePaymentData, authToken);

            if (handlePaymentResult && handlePaymentResult.ok) {
                try {
                    var responseData = JSON.parse(handlePaymentResult.object);

                    if (responseData.checkoutUrl) {
                        handlePaymentStatus.error = false;
                        session.privacy.scalapayAuthURL = handlePaymentStatus.redirectURL = responseData.checkoutUrl;
                        scalapayHelper.setAuthAmount(authAmountMoney.value, basket);
                        scalapayHelper.saveAuthAddressData(basket);
                    } else {
                        Logger.error('ScalaPay payment authorization error. Error: Payment authorization redirect URL not in response');
                    }
                } catch (error) {
                    Logger.error('ScalaPay payment authorization error. Error: {0}', error);
                }
            } else {
                Logger.error('ScalaPay payment authorization error. Error: Service error');
            }
        } else {
            Logger.error('ScalaPay payment authorization error. Error: Authorization token missed');
        }
    } else {
        Logger.error('Scalapay payment authorization error. Error: Scalapay is not applicable');
    }
    if (handlePaymentStatus.error){
        handlePaymentStatus.redirectURL = URLUtils.https('COBilling-Start', 'error', 'payment.not.applicable.error').toString();
    }
    return handlePaymentStatus;
}

function authorize(args) {
    var orderNumber = args.OrderNo,
        scalapayHelper = require('*/cartridge/scripts/helpers/scalapayHelper'),
        OrderMgr = require('dw/order/OrderMgr'),
        URLUtils = require('dw/web/URLUtils'),
        PaymentMgr = require('dw/order/PaymentMgr'),
        order = OrderMgr.getOrder(orderNumber),
        paymentInstrument = args.PaymentInstrument,
        paymentProcessor = PaymentMgr.getPaymentMethod(paymentInstrument.getPaymentMethod()).getPaymentProcessor(),
        isScalapayApplicable = scalapayHelper.isPaymentApplicable(order),
        AuthorizePaymentStatus = {
            error: true,
            redirectURL: URLUtils.https('COBilling-Start', 'error', 'scalapay.errorpayment.error.002').toString()
        };

    Transaction.wrap(function() {
        paymentInstrument.paymentTransaction.setPaymentProcessor(paymentProcessor);
    });

    if (isScalapayApplicable) {
        var authToken = currentSite.getCustomPreferenceValue('scalaPayPaymentToken');

        if (authToken) {
            var paymentToken = scalapayHelper.getPaymentToken(order),
                scalapayAuthorizeService = getScalaPayPaymentAuthorizeService(paymentToken),
                reqData = getAuthorizationReqData(orderNumber, paymentToken),
                isPaymentDataSyncronized = scalapayHelper.isPaymentDataSync(order),
                isAddressDataSyncronized = scalapayHelper.isAddressDataSync(order);

            if (isPaymentDataSyncronized && isAddressDataSyncronized) {
                var authResult = scalapayAuthorizeService.call(reqData, authToken);

                if (authResult && authResult.ok) {
                    AuthorizePaymentStatus.error = false;
                    scalapayHelper.submitOrder(order);
                } else {
                    Logger.error('Scalapay payment capture error. Error: Service error');
                }
            }else{
                Logger.error('Scalapay payment capture error. Error: Order amount and/or order address does not match to authorized values');
            }
        } else {
            Logger.error('Scalapay payment capture error. Error: Authorization token missed');
        }
    } else {
        Logger.error('Scalapay payment capture error. Error: Scalapay is not applicable');
    }
    
    if (AuthorizePaymentStatus.error) {
        session.privacy.scalapayAuthURL = null;
    }

    return AuthorizePaymentStatus;
}

module.exports = {
    Handle: handle,
    Authorize: authorize
};